/* Cache Functions
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CORE_CACHE_H
#define CORE_CACHE_H

#include <inttypes.h>

/** \file cache.h
 * \brief routines for manipulating the cache
 */

/** \brief enable the instruction cache
 * \details unlocks and invalidates the instruction cache and instruction tlb
 *  and activates the instruction cache
 * \note as a side-effect, the branch target buffer is also invalidated
 * \note look at the Intel XScale Core Manual on Page 4.3.2-52 (ICache) and
 *  7.2.2-83 (Control Register) */
EXTERN_C void iCacheEnable();

/** \brief determine wether the instruction cache is activated
 * \return \c TRUE iff cache is enabled
 * \note look at the Intel XScale Core Manual on
 *  Page 7.2.2-83 (Control Register) */
EXTERN_C bool iCacheIsEnabled();

/** \brief disable the instruction cache
 * \details disables the instruction cache, unlocks and invalidates the cache,
 *  branch target buffer and tlb
 * \note look at the Intel XScale Core Manual on Page 4.3.2-52 (ICache)
 *  and 7.2.2-83 (Control Register) */
EXTERN_C void iCacheDisable();

/** \brief invalidate the instruction cache
 * \note as a side effect, the branch target buffer is also invalidated
 * \note this will not invalidate locked regions. If you need this, call
 *  iCacheUnlock() before
 * \note look at the XScale Core Manual on Page 4.3.3-53 (ICache) and
 *  7.2.8-78 (Cache Functions) */
EXTERN_C void iCacheInvalidate();

/** \brief invalidate consecutive lines of the instruction cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \note This will not invalidate the branch target buffer. If you need this,
 *  call btbInvalidate() afterwards
 * \warning the given line should not be locked as this results in an unusable
 *  cache line
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
EXTERN_C void iCacheInvalidateLine(const void * addr, uint32_t len);

/** \brief lock instructions in the instruction cache
 * \param addr base address
 * \param length length
 * \note look at the XScale Core Manual on Page 4.3.4-54 (ICache) and
 *  7.2.10-90 (Cache Lock Down)
 * \note this code must be placed in a non-cacheable region as described in the
 *  Core Manual */
EXTERN_C void iCacheLock(const void * addr, uint32_t len);

/** \brief unlock the instruction cache
 * \note look at the XScale Core Manual on Page 7.2.10-90 (Cache Lock Down) */
EXTERN_C void iCacheUnlock();

/** \brief enable the branch target buffer
 * \details invalidates the buffer and activates it
 * \note look at the XScale Core Manual on Page 5.2.1-59 (BTB) and
 *  7.2.2-83 (Control Register) */
EXTERN_C void btbEnable();

/** \brief determine wether the branch target buffer is enabled
 * \return \c TRUE iff the buffer is enabled
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */
EXTERN_C bool btbIsEnabled();

/** \brief disable the branch target buffer
 * \details invalidates the buffer and deactivates it
 * \note look at the XScale Core Manual on Page 5.2.1-59 (BTB) and
 *  7.2.2-83 (Control Register) */
EXTERN_C void btbDisable();

/** \brief invalidate the branch target buffer
 * \note look at the XScale Core Manual on Page 5.2.2-59 (BTB) and
 *  7.2.8-87 (Cache Functions) */
EXTERN_C void btbInvalidate();

/** \brief enable the data cache.
 * \details unlocks and invalidates the data cache and data tlb,
 *  flushes the write buffer and activates the cache
 * \warning make sure that the MMU is enabled at this point
 * \note look at the XScale Core Manual on Page 6.3.2-69 (DCache) and
 *  7.2.2-83 (Control Register) */
EXTERN_C void dCacheEnable();

/** \brief determine wether the data cache is activated
 * \return \c TRUE iff cache is enabled
 * \note look at the Intel XScale Core Manual on
 *  Page 7.2.2-83 (Control Register) */
EXTERN_C bool dCacheIsEnabled();

/** \brief disable the data cache
 * \details flushes the write buffer, disables the data cache, unlocks and
 *  invalidates the cache and tlb
 * \note look at the Intel XScale Core Manual on Page 6-61 (DCache),
 *  7.2.2-83 (Control Register) */
EXTERN_C void dCacheDisable();

/** \brief unlock the data cache
 * \note look at the XScale Core Manual on Page 7.2.10-90 (Cache Lock Down) */
EXTERN_C void dCacheUnlock();

/** \brief clean the data and mini-data cache
 * \details cleans the data cache by allocating lines in it, then cleans the
 *  mini-data cache by loading abitrary data in it
 * \note look at the Intel XScale Core Manual on Page 6.3.3.1-70 (DCache) */
EXTERN_C void dCacheClean();

/** \brief clean and invalidate the data and mini-data cache
 * \details cleans the data cache by allocating lines in it, then cleans the
 *  mini-data cache by loading abitrary data in it and then invalidates the
 *  caches
 * \note look at the Intel XScale Core Manual on Page 6.3.3.1-70 (DCache) */
EXTERN_C void dCacheCleanAndInvalidate();

/** \brief invalidate the data cache
 * \note this will not invalidate locked regions. If you need this, call
 *  dCacheUnlock() before
 * \note this will not clean the cache. If you need this, call dCacheClean()
 *  before. You can also call dCacheCleanAndInvalidate() for both operations
 * \note look at the Intel XScale Core Manual on Page 6.3.3.1-70 (DCache) */
EXTERN_C void dCacheInvalidate();

/** \brief invalidate consecutive lines of the data cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \warning This will not clean the cache. If you need this, call
 *  dCacheCleanLine() before
 * \warning the given line should not be locked as this results in an unusable
 *  cache line
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
EXTERN_C void dCacheInvalidateLine(const void * addr, uint32_t len);

/** \brief clean consecutive lines of the data cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
EXTERN_C void dCacheCleanLine(const void * addr, uint32_t len);

/** \brief clean and invalidate consecutive lines of the data cache
 * \param addr base address to be cleaned and invalidated
 * \param len length of the area to be cleaned and invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
EXTERN_C void dCacheCleanAndInvalidateLine(const void * addr, uint32_t len);

/** \brief drain write buffer
 * \note look at the XScale Core Manual on Page 6.1.3-64 (DCache)
 *  and 7.2.8-87 (Cache Functions) */
EXTERN_C void dCacheDrainBuffer();

/** \brief invalidate all caches and the branch target buffer
 * \note look at the XScale Core Manual on Page 7.2.8-87 (Cache Functions) */
EXTERN_C void iAndDCacheInvalidate();

#endif

