/* Common Exception handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/
#include <os_arm_bare_registers_def.h>
#include <inttypes.h>
#include <core/mmu.h>
#include <stdio.h>
#include "exception.h"
#include "ledcycle.h"

static const char * getModeLabel(uint32_t psr);

/** \brief some common strings */
const char
	* _EXC_LABEL_ENABLED = "enabled",
	* _EXC_LABEL_DISABLED = "disabled",
	* _EXC_LABEL_VALID = "(valid)",
	* _EXC_LABEL_INVALID = "(invalid)";

/** \brief the RELEASE symbol deactivates the exception handler dynamically.
 * the symbol is marked weak, so it can be overridden by another objectfile
 * when linking the image */
__attribute__((weak))
bool __RELEASE__ = FALSE;

/** \brief initialize the exception handler */
void _excInit() {
	if (__RELEASE__) {
		/* Release: step into ledcycle but don't react to button presses.
		 * -> we never return from here on */
		_excLedCycle(FALSE);
	}
	
	/* (re)initialize uart */
	_excUartInit();
	
	/* disable mmu */
	/* TODO: maybe mmu should be disabled but then SDRAM execution won't work */
	//mmuDisable();
}

/** \brief collect processor status
 * \param stat processor status */
void _excGetProcessorStatus(struct _ExcProcessorStatus * stat) {
	asm volatile(
		/* read cpsr */
		"mrs r0, cpsr\n\t"
		"str r0, %[cpsr]\n\t"
		
		/* read spsr and switch cpsr to spsr, but keep interrupts disabled */
		"mrs ip, spsr\n\t"
		"str ip, %[spsr]\n\t"
		"and r1, ip, #0x1f\n\t"
		"cmp r1, #0x10\n\t"
		"moveq ip, #0x1f\n\t"
		"orr ip, #0xc0\n\t"
		"msr cpsr_c, ip\n\t"
		
		/* now read old sp and lr */
		"mov r1, sp\n\t"
		"mov r2, lr\n\t"
		
		/* restore cpsr */
		"msr cpsr, r0\n\t"
		
		/* save old sp and lr */
		"str r1, %[oldsp]\n\t"
		"str r2, %[oldlr]\n\t"
		
		/* read control register */
		"mrc p15, 0, ip, c1, c0, 0\n\t"
		"str ip, %[creg]\n\t"
		/* read pid register */
		"mrc p15, 0, ip, c13, c0, 0\n\t"
		"mov ip, ip, lsr#25\n\t"
		"str ip, %[pid]\n\t"
		: [spsr] "=m" (stat->spsr),
		  [cpsr] "=m" (stat->cpsr),
		  [oldsp] "=m" (stat->oldsp),
		  [oldlr] "=m" (stat->oldlr),
		  [creg] "=m" (stat->creg),
		  [pid]  "=m" (stat->pid) :: "r0", "r1", "r2", "ip");
	/* translate mode of spsr to a readable string */
	stat->spsrModeLabel = getModeLabel(stat->spsr);
	/* translate mode of cpsr to a readable string */
	stat->cpsrModeLabel = getModeLabel(stat->cpsr);
}

/** \brief get a readable processor mode string from a processor status
 *  (CPSR/SPSR) register
 * \param psr processor status
 * \return processor mode as string */
static const char * getModeLabel(uint32_t psr) {
	switch (psr & 0x1f) {
		case 0x10:
			/* USR Mode */
			return "User";
		case 0x11:
			/* FIQ Mode */
			return "Fast Interrupt";
		case 0x12:
			/* IRQ Mode */
			return "Interrupt";
		case 0x13:
			/* SVC Mode */
			return "Supervisor";
		case 0x15:
			/* Debug Mode */
			return "Debug";
		case 0x17:
			/* ABT Mode */
			return "Abort";
		case 0x1b:
			/* UND Mode */
			return "Undefined";
		case 0x1f:
			/* System mode */
			return "System";
		default:
			return "??";
	}
}

/** \brief get the fault status register
 * \return fault status register (FSR) */
uint32_t _excGetFaultStatus() {
	uint32_t fs;
	asm volatile("mrc p15, 0, %[fs], c5, c0, 0" : [fs] "=r" (fs));
	return fs;
}

/** \brief get the fault address register
 * \return fault address register (FAR) */
uint32_t _excGetFaultAddress() {
	uint32_t far;
	asm volatile("mrc p15, 0, %[far], c6, c0, 0" : [far] "=r" (far));
	return far;
}

/** \brief (re)initialize Standard UART
 * This will set the UART to 115200 8N1 */
void _excUartInit() {
	/* stuart */
	GPIO_SET_ALT_FUNC(46,2,GPIO_IN);
	GPIO_SET_ALT_FUNC(47,1,GPIO_OUT);
	
	/* If sensorboard is used */
	GPIO_SET_ALT_FUNC(81,0,GPIO_OUT)
	GPCR(81) |= GPIO_BIT(81);
	
	/* enable clock */
	CKEN |= CKEN_CKEN5;
	
	STLCR |= LCR_DLAB;  // set uart to initial Baudrate 115200 bps
	STDLL = 0x8;
	STDLH = 0;
	STLCR &= ~(LCR_DLAB);
	// set uart to 8 data bits transmitted in each character
	STLCR = LCR_WLS(3);
	
	/* enable fifo */
	STFCR = FCR_TRFIFOE;
	
	/* enable stuart */
	STIER |= IER_UUE;
}

/** \brief print a status message over UART and terminate it by \r\n
 * \param msg message to be sent */
void _excUartOut(const char * msg) {
	char c;
	uint8_t sent = 0;
	
	/* flush buffer */
	while (!(STLSR & LSR_TEMT));
	
	/* send message */
	while((c = *msg++) != '\0') {
		/* read one character and advance to next position */
		
		/* write one byte */
		STTHR = c;
		++sent;
		
		if (sent >= 64) {
			sent = 0;
			/* flush buffer */
			while (!(STLSR & LSR_TEMT));
		}
	}
}

/** \brief Print collected processor status
 * \param lr Contents of LR register
 * \param fp Contents of FP register
 * \param pstat processor status as filled by _excGetProcessorStatus() */
void _excPrintProcessorStatus(uint32_t lr, const uint32_t * fp,
	const struct _ExcProcessorStatus * pstat) {
	char buf[128];
	
	/* print CPSR, SPSR, LR and FP */
	snprintf(buf, sizeof(buf), " CPSR: 0x%08x  SPSR: 0x%08x       LR:  0x%08x   FP:  0x%08x\n",
		pstat->cpsr, pstat->spsr, lr, (uint32_t)fp);
	_excUartOut(buf);
	
	/* print CR0 and its decoded fields */
	snprintf(buf, sizeof(buf), " CR0:  0x%08x  Endianess: %-10s  EReloc: %-8s  PID: %d\n",
		pstat->creg, (pstat->creg & (1 << 7)) ? "big" : "little",
		(pstat->creg & (1 << 13)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED,
		pstat->pid);
	_excUartOut(buf);
	snprintf(buf, sizeof(buf), " ICache: %-8s  DCache: %-13s  MMU: %-11s  BTB: %s\n",
		(pstat->creg & (1 << 12)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED,
		(pstat->creg & (1 << 2)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED,
		(pstat->creg & (1 << 0)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED,
		(pstat->creg & (1 << 11)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED);
	_excUartOut(buf);
	snprintf(buf, sizeof(buf), " AlignF: %-8s  Rom Prot.: %-10s  System Prot.: %s\n",
		(pstat->creg & (1 << 1)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED,
		(pstat->creg & (1 << 9)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED,
		(pstat->creg & (1 << 8)) ? _EXC_LABEL_ENABLED : _EXC_LABEL_DISABLED);
	_excUartOut(buf);
}



