/* Stack Setup
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/

.include "cpu_internal.s"

/*
  Memory Layout
  |--------------|
  |  FIQ STACK   |
  |--------------|
  |  IRQ STACK   |
  |--------------| <- ...
  |  SVC STACK   |
  |--------------| <- ABT Stack pointer (grows down)
  |  ABT STACK   |
  |--------------| <- UND Stack pointer (grows down)
  |  UND STACK   |
  |--------------| <- this is end + _MALLOC_SIZE (lower bound for UND stack)
  | Dynamic      |
  |  Memory used |
  |  by malloc   |
  |--------------| <- this is, where "end" points
  | .bss section |
  |--------------|
  
  if a specific stack is not used (i.e. its size in the descriptor is zero,
   then the corresponding stack pointer will point to 0x0 to increase chances
   of a fault situation if that stack is ever used unintended
  the stack limit pointer is the same for all modes except FIQ mode, so the
   stack limit points to the lower bound of the UND stack resp. FIQ stack
   see also the gcc option -mapcs-stack-check
*/

.section .loader
/* \param r0 stack end (this is where FIQ sp will point to)
 * expects to run in SVC mode */
.globl _stackSetup
_stackSetup:
	/* since there is no stack yet, we store lr in r3 to return finally */
	mov r3, lr
	
	/* setup FIQ stack */
	ldr r1, =_FIQ_STACK_SIZE
	mov r2, #ARM_CPSR_MODE_FIQ
	bl .Lsetstack
	/* this is a special case, because FIQ mode has its own stack limit */
	mov r2, #ARM_CPSR_MODE_FIQ	@ switch to FIQ mode
	msr cpsr_c, r2
	mov sl, r0					@ set the stack limit to the lower bound
	mov r2, #ARM_CPSR_MODE_SVC	@ switch back
	msr cpsr_c, r2
	
	/* setup IRQ stack */
	ldr r1, =_IRQ_STACK_SIZE
	mov r2, #ARM_CPSR_MODE_IRQ
	bl .Lsetstack
	
	/* setup SVC stack */
	ldr r1, =_SVC_STACK_SIZE
	mov r2, #ARM_CPSR_MODE_SVC
	bl .Lsetstack
	
	/* setup ABT stack */
	ldr r1, =_ABT_STACK_SIZE
	mov r2, #ARM_CPSR_MODE_ABT
	bl .Lsetstack
	
	/* setup UND stack */
	ldr r1, =_UND_STACK_SIZE
	mov r2, #ARM_CPSR_MODE_UND
	bl .Lsetstack
	
	/* set the stack limit for IRQ, SVC, ABT and UND stack */
	mov sl, r0
	
	/* finally return */
	mov pc, r3
	
/* coroutine .Lsetstack:
 * \param r0 upper bound of current stack
 * \param r1 stack size of that mode
 * \param r2 mode
 * \return r0 address for next stack
 * the mode is switched back to SVC to have a common LR register */
.Lsetstack:
	/* switch mode (r2) */
	msr cpsr_c, r2
	/* if the mode is not used, ... */
	cmp r1, #0
	/* then set its stack pointer to zero */
	moveq sp, #0
	/* else to the current address */
	movne sp, r0
	/* recalculate stack address for next stack */
	sub r0, r1
	/* switch back to SVC mode */
	mov r2, #ARM_CPSR_MODE_SVC
	msr cpsr_c, r2
	/* return from coroutine */
	mov pc, lr

