/* Startup Code
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/

.include "cpu_internal.s"

/* ENTRY point of the image:
 * that means that little can be assumed here which is needed for normal
 * program execution, especially for executing code generated by a C compiler.
 *  - there is no stack, which C code relies on
 *  - the data of the program is not copied to its virtual memory address,
 *    where C code expects it to be, i.e. we could not use non-const global or
 *    static variables/data
 *  - the uninitialized data section (bss) is not zeroed, but this is also
 *    expected by C code
 *  - libgcc is not initialized, i.e. no constructors (especially meaningful for
 *    C++ code) have been run
 *  - hardware is not initialized properly, i.e. the core frequency is not set,
 *    caches and the mmu are disabled, etc
 * => thus, the following routines and its helpers are all written in assembler
 *    and do all the mentioned initializations before jumping into the _init
 *    routine of libgcc and finally executing main
 */
.section .loader
.globl _start
_start:
.func _start
	/* switch to SVC mode */
	mov r0, #ARM_CPSR_MODE_SVC
	msr cpsr_c, r0

	/* determine wether we are executed from the location we expected
	 * (i.e. the location given in the linker script) or not (for example
	 * from sdram) */
	
	/* get the start of the image (defined by the linker script) into r0 */
	ldr r0, =__image_start
	/* get the absolute address of symbol start into r1 (should be r0 + 0x20
	 * but we don't rely on that here because we recalculate that below) */
	ldr r1, =_start
	/* get the relative address of symbol start into r2 (this is where we are
	 * executed currently) */
	adr r2, _start
	/* get the relative address of the image start into r1 = r2 - (r1 - r0) */
	sub r1, r0
	rsb r1, r2
	
	teq r0, r1
	/* if r0 and r1 differ, we are not executed from the location we expected,
	 * thus we have to remap some memory using the mmu */
	bne _startRam
	
	/* if r1 and r2 were equal, we are executed from the location we expected,
	 * thus we can do the initializations now and start up */
	
	/* now load the image and start */
	
	/* setup stack for all significant modes according to the symbols provided
	 * by the linker script (see stack.s for more information) */
	ldr r0, =_STACK_END
	bl _stackSetup
	
	/* load all sections of the image to their respective virtual memory
	 * addresses (see loader.s) */
	bl _loadImage
	
	/* enable mmu/caches */
	bl _coreHardwareInitMMU
	
	b _startMain
.endfunc

.globl _startMain
_startMain:
.func _startMain
	bl _coreHardwareClean
	
	/* initialize libgcc if needed (also calls C++ constructors of static
	 * objects) */
	bl _init
	
	/* from now on we can execute all C/C++ code, which was generated by gcc */
	/* initialize hardware (SEnF) */
	bl hardwareInit
	
	/* jump into main */
	mov	r0, #0 @argc
	mov	r1, #0 @argv
	mov r2, #0 @envp
	bl main
	
	/* finish libgcc if needed */
	bl _fini
	
	/* loop forever */
1:
	b 1b
.endfunc


